;;
;; get_argv - C compatible command line parser from the WCRT library.
;;
;; Copyright (c) 2004 by Joergen Ibsen / Jibz
;; All Rights Reserved
;;
;; http://www.ibsensoftware.com/
;;

.386
.model flat,stdcall

option casemap:none

include windows.inc
include kernel32.inc

public get_argv

parse_argv PROTO :DWORD,:DWORD,:DWORD,:DWORD

.code

; =============================================================

get_argv proc uses esi cmdl:DWORD, argc:DWORD

    LOCAL n:DWORD

    mov    eax, cmdl

    test   eax, eax           ; check cmdl != NULL
    jz     return_eax         ;

    xchg   eax, esi           ; esi -> cmdl

    ; get required size of buffer and arg count
    invoke parse_argv, esi, ADDR n, NULL, NULL

    ; allocate memory for buffer
    invoke GlobalAlloc, GMEM_FIXED, eax

    test   eax, eax
    jz     return_eax

    push   eax                ; push address of buffer

    mov    edx, n             ; edx = arg count

    mov    ecx, argc          ; ecx -> argc

    jecxz  @F                 ; store arg count where argc
    mov    [ecx], edx         ; points to (unless it's NULL)
  @@:

    inc    edx                ; room for zero-termination of argv

    lea    ecx, [eax + edx*4] ; ecx -> buffer + size of argv array

    ; parse command line into buffer
    invoke parse_argv, esi, ADDR n, eax, ecx

    pop    eax                ; eax -> argv array

  return_eax:
    ret

get_argv endp

; =============================================================

parse_argv proc uses esi edi cmdl:DWORD, argc:DWORD, argv:DWORD, buffer:DWORD
    ; int parse_argv(const char *cmdl, int *argc,
    ;                char **argv, char *buffer)

    ; NOTE: This function assumes that argv != NULL means that
    ;       buffer != NULL.

    ; NOTE: Defining SIMPLE will remove the code for backslash
    ;       handling. Quoted strings are still handled, and """
    ;       still works to insert a quote.

    xor    edx, edx           ; edx will count the required space

    mov    esi, cmdl
    mov    ecx, argv
    mov    edi, buffer

    push   edx                ; push zero on stack for argc

  skip_ws:
    lodsb

  next_arg:
    test   al, al
    jz     done

    cmp    al, ' '
    jbe    skip_ws

    dec    esi                ; adjust esi so we re-read first non-ws

    xor    eax, eax

    jecxz  @F
    stosb                     ; zero-terminate previous arg
    mov    [ecx], edi         ; store pointer in argv
    add    ecx, 4             ;
  @@:

    inc    dword ptr [esp]    ; increment argc

    add    edx, 5             ; need 5 more bytes

    mov    ah, ' '            ; initialize arg terminator to space

  copy_arg:
    lodsb

    ; quote handling

  quote_check:
    cmp    al, '"'
    jne    bs_check

    xor    ah, ' '            ; toggle terminator between space and 0
    jz     copy_arg           ; if opening quote, continue as normal

    lodsb

    ; backslash handling

  bs_check:

ifndef SIMPLE

    cmp    al, '\'            ; check for '\'
    jne    test_and_write     ;

  bs_parse:
    push   ecx

    xor    ecx, ecx           ; clear counter

  bs_more:
    inc    ecx                ; count number of '\'
    lodsb                     ;
    cmp    al, '\'            ;
    je     bs_more            ;

    dec    esi                ; adjust esi so we re-read first non-'\'

    cmp    al, '"'            ; if next char is not a '"', write all '\'
    jne    bs_write           ;

    shr    ecx, 1             ; half the number of '\'
    jnc    bs_write           ; if uneven we have an escaped '"'

    dec    esi                ; go back to last '\'

  bs_write:
    add    edx, ecx           ; need ecx more bytes

    mov    al, '\'

    cmp    dword ptr [esp], 0
    jz     @F
    rep    stosb              ; store ecx times '\'
  @@:

    pop    ecx

    lodsb

    cmp    al, '\'            ; if not at '\' we need to handle
    jne    quote_check        ; this char normaly

    lodsb                     ; read '"'

endif

    ; check for end of argument and write

  test_and_write:
    cmp    al, ah             ; have we reached end of current arg?
    jbe    next_arg           ;

  write_char:
    jecxz  @F
    stosb                     ; store char in buffer
  @@:

    inc    edx                ; need 1 more byte
    jmp    copy_arg

  done:
    jecxz  @F
    xor    eax, eax
    stosd                     ; zero-terminate last arg
    mov    [ecx], eax         ; zero-terminate argv array
  @@:

    lea    eax, [edx + 16]    ; eax = required space

    mov    ecx, argc

    pop    edx                ; edx = argc

    jecxz  @F
    mov    [ecx], edx         ; store argc
  @@:

    ret

parse_argv endp

; =============================================================

end
